/*
 * xgal.sndsrv.c - ALSA Compatible Sound
 *                 This server is ALSA Specific.
 *
 * Copyright 1994-1995 Sujal M. Patel (smpatel@wam.umd.edu)
 * Copyright 2009 Hans de Goede <hdegoede@redhat.com>
 * Conditions in "copyright.h"
 */

/* $Id: xgal.sndsrv.linux.c,v 1.3 1998/05/11 06:37:20 mrogre Exp $ */
#include <config.h>

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <signal.h>
#include <string.h>
#include <alsa/asoundlib.h>

char *FILENAME[] = {
                     "/explode.raw",
                     "/firetorp.raw",
                     "/shield.raw",
                     "/torphit.raw",
                     "/explode_big.raw",
		     "/ddloo.raw",
		     "/warp.raw",
		     "/smart.raw",
                   };

#define NUM_SOUNDS	(sizeof(FILENAME)/sizeof(char*))

char *sound_buffer[NUM_SOUNDS];
int sound_size[NUM_SOUNDS];

const int fragsize = 256;

/* Terminate: Signal Handler */
void quit ()
{
  exit (0);
}



void init (int argc, char **argv)
{
  int i;
  char s[1024];
  
  if (argc != 3)
  {
    printf ("This program is only executed by xgal\n");
    exit (1);
  }

  for (i=0; i < NUM_SOUNDS; i++)
  {
    s[0] = 0;
    strcat (s, argv[1]);
    if (s[(int)strlen(s) - 1] == '/') FILENAME[i]++;
    strcat (s, FILENAME[i]);
    FILENAME[i] = malloc ((int)strlen (s)+1);
    strcpy (FILENAME[i],s);
    sound_buffer[i]=NULL;
    sound_size[i]=0;
  }

  signal(SIGTERM, quit);  /* Setup Terminate Signal Handler */
}


/*
   Setup PCM:
              Sets speed to 8000hz
              Should set mono mode
              Error checking                
*/
snd_pcm_t *setup_pcm(char *device)
{
  int err;
  snd_pcm_t *handle;

  if ((err = snd_pcm_open(&handle, device, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
    fprintf(stderr, "Playback open error: %s\n", snd_strerror(err));
    return NULL;
  }

  if ((err = snd_pcm_set_params(handle,
              SND_PCM_FORMAT_U8,
              SND_PCM_ACCESS_RW_INTERLEAVED,
              1,
              8000,
              1,
              50000)) < 0) {   /* 0.05sec */
    fprintf(stderr, "Playback open error: %s\n", snd_strerror(err));
    snd_pcm_close(handle);
    return NULL;
  }

  return handle;
}

int read_sound(int k)
{
  int i,fd,size;

  /*fprintf(stderr,"loading sound %d, %s\n",k,FILENAME[k]);*/

  fd = open(FILENAME[k], O_RDONLY);
  if(fd < 0) 
  {
    fprintf (stderr, "xgal.sndsrv: The sound %s could not be opened\n", FILENAME[k]);
    sound_size[k]=-1;
    return(0);
  };
  size=lseek(fd,0,SEEK_END);
  sound_size[k]=(size/fragsize)+1;   /*size in fragments*/
  sound_buffer[k]=malloc(sound_size[k]*fragsize);
  if(sound_buffer[k]==NULL)
  {
    fprintf(stderr,"xgal.sndsrv: couldn't malloc memory for sound\n");
    sound_size[k]=-1;
    close(fd);
    return(0);
  };
  lseek(fd,0,SEEK_SET);
  size = read(fd,sound_buffer[k],size);
  close(fd);
  if (size) {
    for(i=0;i<size;i++)  sound_buffer[k][i]^=0x80;
    memset(sound_buffer[k]+size,0,sound_size[k]*fragsize-size);
  }
  
  /*fprintf(stderr,"sound has been loaded, %d bytes\n",size);*/ /*DEBUG*/
  return(1);
}


void do_everything (snd_pcm_t *handle)
{
  signed char k;
  int i, j ;
  int terminate = -1;             /* Which Sound to Terminate                              */
  int playing[16];                /* Sound numbers that we are playing                     */
  int position[16];		  /* Current position in each sound file */
  int playnum = 0;                /* Number of sounds currently being played               */
  unsigned char final[fragsize];       /* Final Mixing Buffer                                   */
  int premix[fragsize];           /* Use fragsize in case sound card sets it big. -- JEH */
  char *sample;
  snd_pcm_sframes_t frames;
  int first_short_write = 1;

  for(;;)  {
    terminate = -1;
    /* Try to open a new sound if we get an integer on the 'in' pipe */
    i=read(STDIN_FILENO,&k,sizeof(k));
    if(i==0)  {   /* EOF on pipe means parent has closed its end */
        /*fprintf(stderr,"xgal.sndsrv: shutting down\n"); */
        return;
    };
    if(i!=-1)  {  /* there was something in the pipe */
        /*fprintf(stderr,"Just read a %d from pipe\n",(int)k);*/ /*DEBUG*/
        /* Negative means terminate the FIRST sound in the buffer */
        if(k<0)  {
            /*fprintf(stderr,"terminating sound\n");*/ /*DEBUG*/
            terminate = 0;
        } else {
            if(sound_size[(int)k]==0) read_sound(k);
            if(sound_size[(int)k]>0 && playnum<16)  {
	        position[playnum]=0;
                playing[playnum++]=k;
                /*fprintf(stderr,"sound %d added to play queue\n",playnum-1);*/ /*DEBUG*/
            };
        };
    };

    /* terminate a sound if necessary */
    for(i=0;i<playnum;i++)
    {
      if((position[i]==sound_size[playing[i]]) || (terminate==i))
      {
        /*fprintf(stderr,"finished playing sound %d\n",i);*/ /*DEBUG*/
	/*fprintf(stderr,"is was at position %d\n",position[i]);*/ /*DEBUG*/
        memmove(playing+i,playing+i+1,(playnum-i)*sizeof(int));
        memmove(position+i,position+i+1,(playnum-i)*sizeof(int));
        playnum--;i--;
      };
    };

    if(playnum)  {
        /* Mix each sound into the final buffer */
        memset(premix,0,sizeof(premix));
        for(i=0;i<playnum;i++)  {
            sample=sound_buffer[playing[i]]+position[i]*fragsize;
            for(j=0;j<fragsize;j++)  {
                premix[j]+=*(sample+j);
            };
            position[i]++;
        };
        for(i=0;i<fragsize;i++)
            final[i]=(premix[i]>255)?255:(premix[i]<-256?0:(premix[i]>>1)+128);
    } else {
        /* 
           We have no sounds to play
           Just fill the buffer with silence and maybe play it 
        */
        memset(final,128,sizeof(final));
    };
    frames = snd_pcm_writei(handle, final, sizeof(final));
    if (frames < 0)
      frames = snd_pcm_recover(handle, frames, 0);
    if (frames < 0)
      fprintf(stderr, "snd_pcm_writei failed: %s\n", snd_strerror(frames));
    if (frames > 0 && frames < (long)sizeof(final)) {
      if (first_short_write) {
        first_short_write = 0;
        continue;
      }
      fprintf(stderr, "Short write (expected %li, wrote %li)\n", (long)sizeof(final), frames);
    }
    /*
       The sound server is in a tight loop, EXCEPT for this
       write which blocks.  Any optimizations in the above
       code would really be helpful.  Right now the server
       takes up to 7% cpu on a 486DX/50.
    */
  }
}



int main (argc, argv)
int argc;
char **argv;
{
  snd_pcm_t *handle;

  fcntl(STDIN_FILENO,F_SETFL,O_NONBLOCK);
  init (argc, argv);
  handle = setup_pcm(argv[2]);

  if (!handle) exit(1);
  
  do_everything(handle);

  snd_pcm_close(handle);

  return 0;
}
